# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


import functools

from PyQt4.QtGui import QTreeWidget, QTreeWidgetItem

from ...model import observe
from ...plugins import PluginManager, IPluginManager
from ...ui import IDisplay


class ExtensionsViewer(QTreeWidget):
    """ The ExtensionsViewer class is a widget that implements a read-only
    visualisation of all extension points.
    """

    def __init__(self):
        """ Initialise the viewer. """

        super().__init__()

        self.setHeaderLabels(["Identifier", "Contribution"])

        ipluginmanager = IPluginManager(PluginManager().instance)
        self._new_extension_points(ipluginmanager.extension_points)
        observe('extension_points', ipluginmanager,
                self._extension_points_changed)

    def _extension_points_changed(self, change):
        """ Invoked when the list of extension points changes. """

        # At the moment extension points can only be added.
        self._new_extension_points(change.new)

    def _new_extension_points(self, extension_points):
        """ Add a list of new extension points. """

        for extension_point in extension_points:
            itm = QTreeWidgetItem(self, [extension_point.id, ''])

            # Add any existing contributions.
            for contribution in extension_point.contributions:
                self._add_contribution(itm, contribution)

            observe('contributions', extension_point,
                    functools.partial(self._new_contribution, itm))

    def _new_contribution(self, parent, change):
        """ Invoked when a new contribution is made to an extension point. """

        # At the moment contributions can only be added.
        for contribution in change.new:
            self._add_contribution(parent, contribution)

    @staticmethod
    def _add_contribution(parent, contribution):
        """ Add a new contribution. """

        id = getattr(contribution, 'id', '')

        tool_tip = str(contribution)

        idisplay = IDisplay(contribution, exception=False)
        name = '' if idisplay is None else idisplay.name
        if name == '':
            name = tool_tip

        itm = QTreeWidgetItem(parent, [id, name])
        itm.setToolTip(1, tool_tip)
