# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ...model import implements, Model
from ...ui import OptionList, ViewStack, WizardPage

from .. import IIoManager, IIoManagerUi, IStorage

from .location_wizard import LocationWizard


@implements(IIoManagerUi)
class IoManagerUi(Model):
    """ The IoManagerUi class implements the various storage-independent user
    interfaces.
    """

    # FIXME: The newlines.
    # The text of the title used in the wizard pages asking the user to choose
    # a storage location.
    choose_location_title = "Choose a storage location"

    # The text of the sub-title used in the wizard page asking the user to
    # choose a readable storage location.
    choose_readable_location_subtitle = "Enter the storage location where you want to read from.\n"

    # The text of the sub-title used in the wizard page asking the user to
    # choose some readable storage.
    choose_readable_storage_subtitle = "From the list below choose where you want to read from.\n"

    # The text of the title used in the wizard pages asking the user to choose
    # some storage.
    choose_storage_title = "Choose the storage"

    # The text of the sub-title used in the wizard page asking the user to
    # choose a writeable storage location.
    choose_writeable_location_subtitle = "Enter the storage location where you want to write to.\n"

    # The text of the sub-title used in the wizard page asking the user to
    # choose some writeable storage.
    choose_writeable_storage_subtitle = "From the list below choose where you want to write to.\n"

    def readable_location_wizard_page(self, id=''):
        """ Create a wizard page factory for asking the user to provide a
        readable storage location.

        :param id:
            is the optional identifier of the page.
        :return:
            the wizard page factory.  The page's view will be an implementation
            of :class:`~dip.ui.IViewStack`.
        """

        return self._location_page(id, self.choose_readable_location_subtitle)

    def readable_storage_wizard_page(self, bind_to, storage_list=None, id=''):
        """ Create a wizard page factory for asking the user to select from a
        list of readable storage.

        :param bind_to:
            is the attribute of the model containing the selected storage.
        :param storage_list:
            is the optional list of available readable storage.
        :param id:
            is the optional identifier of the page.
        :return:
            the wizard page factory.
        """

        return self._storage_page(bind_to, storage_list, id,
                self.choose_readable_storage_subtitle)

    def readable_storage_location(self, window_title, storage_list=None, default_location=None, hints=None, parent=None):
        """ Ask the user to provide a readable storage location.

        :param window_title:
            is the window title to use in any wizards and dialogs.
        :param storage_list:
            is the storage from which the use may choose.  If it is not
            specified then a list of all available readable storage is used.
        :param default_location:
            is the optional default location.
        :param hints:
            is an optional source of hints.
        :param parent:
            is the optional parent view.
        :return:
            the storage location or ``None`` if the user declined to provide
            one.
        """

        if storage_list is None:
            iiomanager = IIoManager(self.io_manager)

            storage_list = [storage for storage in iiomanager.storage
                    if IStorage(storage).readable]

        # See if a wizard is needed.
        if len(storage_list) == 1:
            # A single storage type will never have implicit locations.
            storage = storage_list[0]

            location = storage.ui.get_read_location(window_title,
                    default_location=default_location, hints=hints,
                    parent=parent)
        else:
            wizard = LocationWizard(ui=self,
                    storage_list=storage_list, location=default_location,
                    hints=hints, window_title=window_title)

            location = wizard.location if wizard.execute(parent) else None

        return location

    def writeable_location_wizard_page(self, id=''):
        """ Create a wizard page factory for asking the user to provide a
        writeable storage location.

        :param id:
            is the optional identifier of the page.
        :return:
            the wizard page factory.  The page's view will be an implementation
            of :class:`~dip.ui.IViewStack`.
        """

        return self._location_page(id, self.choose_writeable_location_subtitle)

    def writeable_storage_wizard_page(self, bind_to, storage_list=None, id=''):
        """ Create a wizard page factory for asking the user to select from a
        list of writeable storage.

        :param bind_to:
            is the attribute of the model containing the selected storage.
        :param storage_list:
            is the optional list of available writeable storage.
        :param id:
            is the optional identifier of the page.
        :return:
            the wizard page factory.
        """

        return self._storage_page(bind_to, storage_list, id,
                self.choose_writeable_storage_subtitle)

    def writeable_storage_location(self, window_title, storage_list=None, default_location=None, hints=None, parent=None):
        """ Ask the user to provide a writeable storage location.

        :param window_title:
            is the window title to use in any wizards and dialogs.
        :param storage_list:
            is the storage from which the use may choose.  If it is not
            specified then a list of all available writeable storage is used.
        :param default_location:
            is the optional default location.
        :param hints:
            is an optional source of hints.
        :param parent:
            is the optional parent view.
        :return:
            the storage location or ``None`` if the user declined to provide
            one.
        """

        if storage_list is None:
            iiomanager = IIoManager(self.io_manager)

            storage_list = [storage for storage in iiomanager.storage
                    if IStorage(storage).writeable]

        # See if a wizard is needed.
        if len(storage_list) == 1:
            # A single storage type will never have implicit locations.
            storage = storage_list[0]

            location = storage.ui.get_write_location(window_title,
                    default_location=default_location, hints=hints,
                    parent=parent)
        else:
            wizard = LocationWizard(ui=self, reading=False,
                    storage_list=storage_list, location=default_location,
                    hints=hints, window_title=window_title)

            location = wizard.location if wizard.execute(parent) else None

        return location

    def _location_page(self, id, subtitle):
        """ Return an appropriately configured location wizard page factory.
        """

        return WizardPage(ViewStack(), auto_form='never', id=id,
                title=self.choose_location_title, subtitle=subtitle)

    def _storage_page(self, bind_to, storage_list, id, subtitle):
        """ Return an appropriately configured storage wizard page factory. """

        if storage_list is None:
            storage_list = []

        return WizardPage(
                OptionList(bind_to, options=storage_list, allow_none=False,
                        sorted=True),
                auto_form='never', id=id, title=self.choose_storage_title,
                subtitle=subtitle)
