# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Adapter, implements, Model, observe
from ..ui import IController, IDisplay, IView, IViewBinding

from .i_close_view_veto import ICloseViewVeto
from .i_managed_model import IManagedModel
from .i_model_manager_tool import IModelManagerTool
from .i_model_publisher import IModelPublisher
from .i_open_model import IOpenModel
from .i_shell import IShell


@implements(ICloseViewVeto, IModelPublisher, IOpenModel)
class BaseManagedModelTool(Model):
    """ The BaseManagedModelTool class is a base class for :term:`tools<tool>`
    that handle :term:`managed models<managed model>`.  It automatically manages
    the different lists of views and provides a convenient interface to the
    changing of the active view and model.
    """

    def create_views(self, model):
        """ Create the views for a managed model.

        This must be implemented by a sub-class.

        :param model:
            is the model.
        :return:
            the list of views.
        """

        raise NotImplementedError

    def veto(self, view):
        """ Determine if the view is to be prevented from being closed.

        :param view:
            is the view.
        :return:
            ``True`` if the close of the view is to be prevented.
        """

        return IModelManagerTool(self.manager).veto_close_view(view, self)

    def open_model(self, location):
        """ Open a model at a storage location.

        :param location:
            is the storage location.
        """

        IModelManagerTool(self.manager).open_model(location, self)

    @observe('models')
    def _models_changed(self, change):
        """ Invoked when the list of the tool's models changes. """

        ishell = IShell(self.shell)

        # This will take care of the same model being handled by multiple tools.
        for model in change.old:
            imanagedmodel = IManagedModel(model)

            observe('name', IDisplay(model), self._model_name_changed,
                    remove=True)

            for view in imanagedmodel.views:
                # Remove from the tool before removing from the shell.
                self.views.remove(view)
                ishell.views.remove(view)

            imanagedmodel.views = []

        for model in change.new:
            imanagedmodel = IManagedModel(model)

            views = self.create_views(model)

            # Each view takes the name of the model.
            idisplay = IDisplay(model)

            for view in views:
                # Make sure all views implement IView at least.
                IView(view)

                view_idisplay = IDisplay(view)

                # Set the short name first as it is less likely to be being
                # observed.
                view_idisplay.short_name = idisplay.short_name
                view_idisplay.name = idisplay.name

                # Track the view's valid state.
                controller = IViewBinding(view).controller
                if controller is not None:
                    observe('invalid_reason', IController(controller),
                            self._invalid_changed)

            imanagedmodel.views.extend(views)
            self.views.extend(views)
            ishell.views.extend(views)

            observe('name', idisplay, self._model_name_changed)

    @observe('current_view')
    def _current_view_changed(self, change):
        """ Invoked when the tool's current view changes. """

        view = change.new

        # Find the model.
        for model in self.models:
            if view in IManagedModel(model).views:
                # Don't publish an invalid model.
                controller = IViewBinding(view).controller
                if controller is not None:
                    if IController(controller).invalid_reason != '':
                        model = None

                break
        else:
            model = None

        self.published_model = model

    def _invalid_changed(self, change):
        """ Invoked when a view's invalid state changes. """

        # Publish the model if the controller says the view is valid.
        self.published_model = change.model.model if change.new == '' else None

    def _model_name_changed(self, change):
        """ Invoked when the name of a model changes. """

        # This is a bit messy.  We want to get the underlying model but the
        # model in the change may be it or something that has adapted it.
        idisplay = change.model
        model = idisplay.adaptee if isinstance(idisplay, Adapter) else idisplay

        for view in IManagedModel(model).views:
            view_idisplay = IDisplay(view)

            view_idisplay.short_name = idisplay.short_name
            view_idisplay.name = idisplay.name
