# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ....model import implements, Model, observe

from ... import IDirty, ITool, IShell


@implements(ITool)
class DirtyTool(Model):
    """ The DirtyTool is the default implementation of a :term:`tool`
    that supports the toolkit specific method of indicating that a shell has
    unsaved data.
    """

    # The tool's identifier.
    id = 'dip.shell.tools.dirty'

    @observe('shell')
    def _shell_changed(self, change):
        """ Invoked when the shell changes. """

        if change.old is not None:
            ishell = IShell(change.old)

            observe('tools', ishell, self._tools_changed, remove=True)

            for tool in ishell.tools:
                self._observe_tool(tool, remove=True)

        if change.new is not None:
            ishell = IShell(change.new)

            for tool in ishell.tools:
                self._observe_tool(tool)

            observe('tools', ishell, self._tools_changed)

    def _tools_changed(self, change):
        """ Invoked when the list of tools changes. """

        for tool in change.old:
            self._observe_tool(tool, remove=True)

        for tool in change.new:
            self._observe_tool(tool)

    def _observe_tool(self, tool, remove=False):
        """ Observe (or not) the dirty sate of a tool. """

        idirty = IDirty(tool, exception=False)
        if idirty is not None:
            observe('dirty', idirty, self._tool_dirty_changed, remove=remove)

    def _tool_dirty_changed(self, change):
        """ Invoked when a tool's dirty state changes. """

        # Handle the trivial case where the shell doesn't support the feature.
        shell_idirty = IDirty(self.shell, exception=False)
        if shell_idirty is None:
            return

        is_dirty = False

        for tool in IShell(self.shell).tools:
            idirty = IDirty(tool, exception=False)
            if idirty is not None and idirty.dirty:
                is_dirty = True
                break

        shell_idirty.dirty = is_dirty
