# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from PyQt4.QtGui import QAbstractButton, QLayout, QLineEdit

from ....io import IFilterHints, IoManager, IIoManagerUi, IStorageLocation
from ....model import adapt, DelegatedTo, Instance, Model
from ....ui import IStorageLocationEditor

from ..editor_layout_adapter import EditorLayoutAdapter
from ..visual_items import find_visual_item


class StorageLocationEditorComponents(Model):
    """ The StorageLocationEditorComponents class is an internal mixin for
    adapters that access the embedded components of a storage location editor.
    """

    # The line edit.
    line_edit = Instance(QLineEdit)

    @line_edit.default
    def line_edit(self):
        """ Invoked to return the tree widget. """

        return find_visual_item(self.adaptee, QLineEdit)

    # The browse button.
    browse_button = Instance(QAbstractButton)

    @browse_button.default
    def browse_button(self):
        """ Invoked to return the add button. """

        return find_visual_item(self.adaptee, QAbstractButton)

    @classmethod
    def usable_editor(cls, item):
        """ Return True if an item can be used as an editor. """

        # The line edit must be an immediate child otherwise the criteria would
        # be far too wide.
        if find_visual_item(item, QLineEdit, recurse=False) is None:
            return False

        if find_visual_item(item, QAbstractButton) is None:
            return False

        return True


@adapt(QLayout, to=[IStorageLocationEditor, IFilterHints])
class QLayoutIStorageLocationEditorAdapter(EditorLayoutAdapter, StorageLocationEditorComponents):
    """ An adapter to implement IStorageLocationEditor for a storage location
    editor that embeds a :class:`~PyQt4.QtGui.QLineEdit` and a
    :class:`~PyQt4.QtGui.QAbstractButton` that implements a browse button.
    """

    # The filter hints.
    filter = DelegatedTo('filter_hints')

    # Remember the storage location between clicks of the browse button.
    _default_location = Instance(IStorageLocation)

    @classmethod
    def isadaptable(cls, adaptee):
        """ Determine if this adapter can adapt a particular layout.

        :param adaptee:
            is the layout.
        :return:
            ``True`` if the layout can be adapted.
        """

        return cls.usable_editor(adaptee)

    def configure(self, properties):
        """ Configure the editor. """

        line_edit = self.line_edit

        # For automation purposes give the line edit an the id if it doesn't
        # already have one.
        if line_edit.objectName() == '':
            line_edit.setObjectName(self.id)

        # Connect things up.
        line_edit.textEdited.connect(self.qt4_notify)
        self.browse_button.clicked.connect(self._on_clicked)

        # Apply any properties to the line edit.
        try:
            line_edit.pyqtConfigure(**properties)
        except AttributeError:
            pass

    @IStorageLocationEditor.read_only.getter
    def read_only(self):
        """ Invoked to get the read-only state. """

        return self.line_edit.isReadOnly()

    @read_only.setter
    def read_only(self, value):
        """ Invoked to set the read-only state. """

        self.line_edit.setReadOnly(value)
        self.browse_button.setVisible(not value)

    @IStorageLocationEditor.value.getter
    def value(self):
        """ Invoked to get the editor's value. """

        return self.line_edit.text()

    @value.setter
    def value(self, value):
        """ Invoked to set the editor's value. """

        # Save for the notifier.
        self.qt4_old_value = value

        self.line_edit.setText(value)

    @IStorageLocationEditor.invalid_reason.getter
    def invalid_reason(self):
        """ Invoked to get the invalid reason. """

        text = self.line_edit.text()

        if text == '':
            invalid_reason = "a storage location is required" if self.required else ''
        else:
            if self.mode == 'open':
                locations = IoManager.readable_locations_from_string(
                        self.format, text)
            else:
                locations = IoManager.writeable_locations_from_string(
                        self.format, text)

            nr_locations = len(locations)
            if nr_locations == 0:
                invalid_reason = "a valid {0} storage location is required".format("readable" if self.mode == 'open' else "writeable")
            elif nr_locations > 1:
                invalid_reason = "an unambiguous storage location is required"
            else:
                invalid_reason = ''

        return invalid_reason

    def set_focus(self):
        """ Give the focus to the line edit. """

        self.line_edit.setFocus()

    def _on_clicked(self):
        """ Invoked when the browse button is clicked. """

        iiomanagerui = IIoManagerUi(IoManager.ui)

        if self.mode == 'open':
            location = iiomanagerui.readable_storage_location(
                    self.window_title if self.window_title != '' else "Open",
                    default_location=self._default_location, hints=self,
                    parent=self.browse_button)
        else:
            location = iiomanagerui.writeable_storage_location(
                    self.window_title if self.window_title != '' else "Save",
                    default_location=self._default_location, hints=self,
                    parent=self.browse_button)

        self.line_edit.activateWindow()
        self.line_edit.setFocus()

        if location is not None:
            self._default_location = location
            self.value = str(location)
