# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


import io
import os.path

from PyQt4.QtCore import QFile, QIODevice, Qt, QTimer
from PyQt4.QtGui import (QAction, QApplication, QCheckBox, QComboBox, QDialog,
        QDialogButtonBox, QDockWidget, QDoubleSpinBox, QFileDialog,
        QFormLayout, QGroupBox, QHBoxLayout, QKeySequence, QLabel, QLayout,
        QLineEdit, QListWidget, QMainWindow, QMessageBox, QMenu, QMenuBar,
        QPushButton, QSpinBox, QSplitter, QStackedLayout, QTabWidget,
        QTextEdit, QToolButton, QVBoxLayout, QWhatsThis, QWidget, QWizard,
        QWizardPage)
from PyQt4.uic import loadUiType

from ...model import implements, Instance, Tuple

from ..i_toolkit import IToolkit
from ..base_toolkit import BaseToolkit

from .utils import as_QLayout, as_QWidget, as_QWidget_parent
from .widgets import (DipFilesystemLocationEditor, DipListEditor,
        DipRadioButtons, DipStorageLocationEditor, DipTableEditor)


@implements(IToolkit)
class Toolkit(BaseToolkit):
    """ The Toolkit class implements the :term:`toolkit` that uses the Qt4
    GUI library.
    """

    # A map of declarative factories to view types.
    view_type_to_declarative_factory_map = Tuple()

    def action(self, parent):
        """ Create an action.

        :param parent:
            is the optional parent view.
        :return:
            the action which will be a :class:`~PyQt4.QtGui.QAction`.
        """

        return QAction(parent)

    def application(self, argv, **properties):
        """ Create a singleton application.

        :param argv:
            is the sequence of command line arguments.
        :param \*\*properties:
            are the keyword arguments used as toolkit specific property names
            and values that are used to configure the application.
        :return:
            the application.  Repeated calls must return the same application.
        """

        app = QApplication.instance()

        if app is None:
            app = QApplication(argv, **properties)

        return app

    def call_after(self, callable, after):
        """ Call a callable after a delay.

        :param callable:
            is the callable.
        :param after:
           is the delay in milliseconds.  If it is zero then the callable is
           called in the next iteration of the event loop.
        """

        QTimer.singleShot(after, callable)

    def check_box(self, parent):
        """ Create an editor based on a check box.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~PyQt4.QtGui.QCheckBox`.
        """

        return QCheckBox(as_QWidget_parent(parent))

    def close_action(self):
        """ Create an action to handle closing an object.

        :return:
            the action.
        """

        return QAction("&Close", None, shortcut=QKeySequence.Close)

    def combo_box(self, parent):
        """ Create an editor based on a combo box.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~PyQt4.QtGui.QComboBox`.
        """

        return QComboBox(as_QWidget_parent(parent))

    def dialog(self, parent):
        """ Create a dialog view.

        :param parent:
            is the optional parent view.
        :return:
            the view which will be an instance of
            :class:`~PyQt4.QtGui.QDialog`.
        """

        dialog = QDialog(as_QWidget_parent(parent))

        layout = QVBoxLayout()
        layout.addWidget(QDialogButtonBox())
        dialog.setLayout(layout)

        return dialog

    def displayable(self, view):
        """ Return a :term:`view` that can be displayed as a top-level window.

        :param view:
            is the view to be displayed.
        :return:
            the view, possibly wrapped so that it can be displayed.
        """

        return as_QWidget(view)

    def dock(self, parent):
        """ Create a dock view.

        :param parent:
            is the optional parent view.
        :return:
            the view which will be an instance of
            :class:`~PyQt4.QtGui.QDockWidget`.
        """

        return QDockWidget(as_QWidget_parent(parent))

    def declarative_factory_for_view(self, view):
        """ Get an :class:`~dip.ui.EditorFactory` type object that will create
        a factory that creates views of a particular type.

        :param view:
            is the view.
        :return:
            the :class:`~dip.ui.EditorFactory` sub-class.
        """

        for vt, df in self.view_type_to_declarative_factory_map:
            if isinstance(view, vt):
                return df

        raise TypeError(
                "there is no editor factory that creates instances of "
                "{0}".format(type(view).__name__))

    @view_type_to_declarative_factory_map.default
    def view_type_to_declarative_factory_map(self):
        """ Invoked to get the default map between view types and declarative
        factories.
        """

        from ... import ui

        return (
            (QCheckBox,      ui.CheckBox),
            (QComboBox,      ui.ComboBox),
            (QDoubleSpinBox, ui.FloatSpinBox),
            (QLineEdit,      ui.LineEditor),
            (DipListEditor,  ui.ListEditor),
            (QPushButton,    ui.PushButton),
            (QSpinBox,       ui.SpinBox),
            (QTabWidget,     ui.TabBar),
            (DipTableEditor, ui.TableEditor),
            (QTextEdit,      ui.TextEditor),
            (QToolButton,    ui.ToolButton)
        )

    def designer_load(self, ui_file_name):
        """ Load a file containing a user interface design.

        :param ui_file_name:
            is the name of the file.
        :return:
            an opaque object that will be passed to
            :meth:`~dip.toolkits.IToolkit.designer_view`.
        """

        # We use QFile in order to support Qt resource files.
        ui_file = QFile(ui_file_name)

        if not ui_file.open(QIODevice.ReadOnly):
            raise IOError("unable to open '{0}'".format(ui_file_name))

        ui_xml = io.BytesIO(ui_file.readAll())

        ui_file.close()

        ui, base = loadUiType(ui_xml)

        # Return the UI "painter" and the base Qt type to "paint".
        return ui(), base

    def designer_view(self, design, parent):
        """ Create a view from a previously loaded user interface design.

        :param design:
            is an opaque object returned by a previous call to
            :meth:`~dip.toolkits.IToolkit.designer_load`.
        :param parent:
            is the optional parent view.
        :return:
            the view.
        """

        ui, base = design

        # Create an instance of the base class and "paint" it.
        view = base(parent)
        ui.setupUi(view)

        # Adapt the view to something specific if possible.

        from ... import ui

        if isinstance(view, QDialog):
            interface = ui.IDialog
        elif isinstance(view, QMainWindow):
            interface = ui.IMainWindow
        elif isinstance(view, QWizard):
            interface = ui.IWizard
        else:
            interface = ui.IView

        interface(view)

        return view

    def filesystem_location_editor(self, parent):
        """ Create an editor that implements a filesystem file/directory
        location editor.

        :param parent:
            is the optional parent view.
        :return:
            the editor.  Any factory properties will be applied to an internal
            :class:`~PyQt4.QtGui.QLineEdit`.
        """

        return DipFilesystemLocationEditor(as_QWidget_parent(parent))

    def find_view(self, root, id):
        """ Find the view with a particular identifier.

        :param root:
            is the root view to begin the search.
        :param id:
            is the identifier of the view to search for.
        :return:
            the view.  An exception will be raised if the view could not be
            found.
        """

        # FIXME: What if root, or the target widget is a QLayout?
        if root.objectName() == id:
            return root

        widget = root.findChild(QWidget, id)

        if widget is None:
            raise NameError("'{0}' has no QLayout or QWidget with an object "
                    "name of '{1}'".format(root, id))

        return widget

    def float_spin_box(self, parent):
        """ Create an editor for floats based on a spin box.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~PyQt4.QtGui.QDoubleSpinBox`.
        """

        return QDoubleSpinBox(as_QWidget_parent(parent))

    def form(self, parent):
        """ Create a layout to contain a number of items arranged as a form.

        :param parent:
            is the optional parent view.
        :return:
            the layout which will be an instance of
            :class:`~PyQt4.QtGui.QFormLayout`.
        """

        return QFormLayout()

    def get_open_file(self, window_title, directory='', filter='', parent=None):
        """ Get the name of a file to open from the user.

        :param window_title:
            is the window title, typically used as the title of a dialog.
        :param directory:
            is the name of an optional initial directory or file.
        :param filter:
            is the optional file filter.
        :param parent:
            is the optional parent view.
        :return:
            the name of the file to open or an empty string if there was none.
        """

        return QFileDialog.getOpenFileName(as_QWidget_parent(parent),
                window_title, directory, filter)

    def get_save_file(self, window_title, directory='', filter='', parent=None):
        """ Get the name of a file to save from the user.

        :param window_title:
            is the window title, typically used as the title of a dialog.
        :param directory:
            is the name of an optional initial directory or file.
        :param filter:
            is the optional file filter.
        :param parent:
            is the optional parent view.
        :return:
            the name of the file to save or an empty string if there was none.
        """

        return QFileDialog.getSaveFileName(as_QWidget_parent(parent),
                window_title, directory, filter)

    def group_box(self, parent):
        """ Create a group box view.

        :param parent:
            is the optional parent view.
        :return:
            the view which will be an instance of
            :class:`~PyQt.QtGui.QGroupBox`.
        """

        return QGroupBox(as_QWidget_parent(parent))

    def h_box(self, parent):
        """ Create a layout to contain a number of items arranged horizontally.

        :param parent:
            is the optional parent view.
        :return:
            the layout which will be an instance of
            :class:`~PyQt4.QtGui.QHBoxLayout`.
        """

        return QHBoxLayout()

    def information(self, window_title, text, parent):
        """ Display a informational message to the user.

        :param window_title:
            is the window title, typically used as the title of a dialog.
        :param text:
            is the text of the message.
        :param parent:
            is the optional parent view.
        """

        QMessageBox.information(as_QWidget_parent(parent), window_title, text)

    def label(self, parent):
        """ Create an editor for displaying a string.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~PyQt4.QtGui.QLabel`.
        """

        return QLabel(as_QWidget_parent(parent))

    def line_editor(self, parent):
        """ Create an editor based on a line editor.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~PyQt4.QtGui.QLineEdit`.
        """

        return QLineEdit(as_QWidget_parent(parent))

    def list_editor(self, parent):
        """ Create an editor for a list.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~dip.toolkits.qt4.widgets.DipListEditor`.  Any factory
            properties will be applied to an internal
            :class:`~PyQt4.QtGui.QTreeWidget`.
        """

        return DipListEditor(as_QWidget_parent(parent))

    def main_window(self, parent):
        """ Create a main window view.

        :param parent:
            is the optional parent view.
        :return:
            the view which will be an instance of
            :class:`~PyQt4.QtGui.QMainWindow`.
        """

        return QMainWindow(as_QWidget_parent(parent))

    def menu(self, parent):
        """ Create a menu.

        :param parent:
            is the optional parent view.
        :return:
            the menu which will be an instance of :class:`~PyQt4.QtGui.QMenu`.
        """

        return QMenu(as_QWidget_parent(parent))

    def menu_bar(self, parent):
        """ Create a menu bar.

        :param parent:
            is the optional parent view.
        :return:
            the menu bar which will be an instance of
            :class:`~PyQt4.QtGui.QMenuBar`.
        """

        # Ignore the parent for OS/X.
        return QMenuBar()

    def message_area(self, parent):
        """ Create a message area.

        :param parent:
            is the optional parent view.
        :return:
            the message area which will be an instance of
            :class:`~PyQt4.QtGui.QLabel`.
        """

        return QLabel(as_QWidget_parent(parent), styleSheet='color: red',
                wordWrap=True)

    def new_action(self):
        """ Create an action to handle creating a new object.

        :return:
            the action.
        """

        return QAction("&New...", None, shortcut=QKeySequence.New)

    def open_action(self):
        """ Create an action to handle opening an object.

        :return:
            the action.
        """

        return QAction("&Open...", None, shortcut=QKeySequence.Open)

    def option_list(self, parent):
        """ Create an editor used to select from a sequence of options.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~PyQt4.QtGui.QListWidget`.
        """

        return QListWidget(as_QWidget_parent(parent))

    def push_button(self, parent):
        """ Create an editor based on a push button.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~PyQt4.QtGui.QPushButton`.
        """

        return QPushButton(as_QWidget_parent(parent))

    # The map of dip message box buttons to the Qt equivalents.
    _button_map = {'yes': QMessageBox.Yes, 'no': QMessageBox.No,
            'cancel': QMessageBox.Cancel, 'save': QMessageBox.Save,
            'discard': QMessageBox.Discard}

    @classmethod
    def _qbutton_from_button(cls, button):
        """ Return the Qt standard button from the given button name. """

        try:
            return cls._button_map[button]
        except KeyError:
            raise ValueError("invalid button '{0}'".format(button))

    def question(self, window_title, text, parent, detail, buttons):
        """ Ask the user a question.

        :param window_title:
            is the window title, typically used as the title of a dialog.
        :param text:
            is the text of the question.
        :param parent:
            is the optional parent view.
        :param detail:
            is the optional additional detail.
        :param buttons:
            is the sequence of buttons to display.  Possible buttons are 'yes',
            'no', 'cancel', 'save' and 'discard'.  The first in the sequence is
            used as the default.
        :return:
            The button that was pressed.
        """

        parent = as_QWidget_parent(parent)

        # Configure the buttons.
        qbuttons = 0
        qdefault = 0
        for button in buttons:
            qbutton = self._qbutton_from_button(button)
            qbuttons |= qbutton

            if qdefault == 0:
                qdefault = qbutton

        if detail != '':
            msg_box = QMessageBox(QMessageBox.Question, window_title, text,
                    qbuttons, parent, detailedText=detail)
            msg_box.setDefaultButton(qdefault)
            answer = msg_box.exec()
        else:
            answer = QMessageBox.question(parent, window_title, text, qbuttons,
                    qdefault)

        for button, qbutton in self._button_map.items():
            if qbutton == answer:
                return button

        # This should never happen.
        raise ValueError("unexpected Qt button: {0}".format(answer))

    def quit_action(self):
        """ Create an action to handle quitting the application.

        :return:
            the action.
        """

        return QAction("&Quit", None, shortcut=QKeySequence.Quit)

    def radio_buttons(self, parent):
        """ Create an editor that implements a set of radio buttons.

        :param parent:
            is the optional parent view.
        :return:
            the editor.  Any factory properties will be applied to each of the
            buttons.
        """

        return DipRadioButtons(as_QWidget_parent(parent))

    def save_action(self):
        """ Create an action to handle saving an object.

        :return:
            the action.
        """

        return QAction("&Save", None, shortcut=QKeySequence.Save)

    def save_as_action(self):
        """ Create an action to handle saving an object under a new name.

        :return:
            the action.
        """

        return QAction("Save &As...", None, shortcut=QKeySequence.SaveAs)

    def spin_box(self, parent):
        """ Create an editor based on a spin box.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~PyQt4.QtGui.QSpinBox`.
        """

        return QSpinBox(as_QWidget_parent(parent))

    def splitter(self, parent):
        """ Create a splitter view to contain a number of items.

        :param parent:
            is the optional parent view.
        :return:
            the view which will be an instance of
            :class:`~PyQt4.QtGui.QSplitter`.
        """

        return QSplitter(as_QWidget_parent(parent))

    def storage_location_editor(self, parent):
        """ Create an editor that implements a storage location editor.

        :param parent:
            is the optional parent view.
        :return:
            the editor.  Any factory properties will be applied to an internal
            :class:`~PyQt4.QtGui.QLineEdit`.
        """

        return DipStorageLocationEditor(as_QWidget_parent(parent))

    def tab_bar(self, parent):
        """ Create a tab bar view.

        :param parent:
            is the optional parent view.
        :return:
            the view which will be an instance of
            :class:`~PyQt4.QtGui.QTabWidget`.
        """

        return QTabWidget(as_QWidget_parent(parent))

    def table_editor(self, parent):
        """ Create an editor for a table.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~dip.toolkits.qt4.widgets.DipTableEditor`.  Any factory
            properties will be applied to an internal
            :class:`~PyQt4.QtGui.QTableWidget`.
        """

        return DipTableEditor(as_QWidget_parent(parent))

    def text_editor(self, parent):
        """ Create an editor based on a text editor.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~PyQt4.QtGui.QTextEdit`.
        """

        return QTextEdit(as_QWidget_parent(parent))

    def tool_button(self, parent):
        """ Create an editor based on a tool button.

        :param parent:
            is the optional parent view.
        :return:
            the editor which will be an instance of
            :class:`~PyQt4.QtGui.QToolButton`.
        """

        return QToolButton(as_QWidget_parent(parent))

    def v_box(self, parent):
        """ Create a layout to contain a number of items arranged vertically.

        :param parent:
            is the optional parent view.
        :return:
            the layout which will be an instance of
            :class:`~PyQt4.QtGui.QVBoxLayout`.
        """

        return QVBoxLayout()

    def view_stack(self, parent):
        """ Create a layout to contain a stack of items such that only one is
        visible at a time.

        :param parent:
            is the optional parent view.
        :return:
            the layout which will be an instance of
            :class:`~PyQt4.QtGui.QStackedLayout`.
        """

        return QStackedLayout()

    def warning(self, window_title, text, parent, detail):
        """ Display a warning message to the user.

        :param window_title:
            is the window title, typically used as the title of a dialog.
        :param text:
            is the text of the warning.
        :param parent:
            is the optional parent view.
        :param detail:
            is the optional additional detail.
        """

        parent = as_QWidget_parent(parent)

        if detail != '':
            msg_box = QMessageBox(QMessageBox.Warning, window_title, text,
                    parent=parent, detailedText=detail)
            msg_box.exec()
        else:
            QMessageBox.warning(parent, window_title, text)

    def whats_this_action(self):
        """ Create an action to handle "What's This?".

        :return:
            the action.
        """

        return QWhatsThis.createAction()

    def wizard(self, parent):
        """ Create a wizard view.

        :param parent:
            is the optional parent view.
        :return:
            the view which will be an instance of
            :class:`~PyQt4.QtGui.QWizard`.
        """

        return QWizard(as_QWidget_parent(parent))

    def wizard_page(self, parent):
        """ Create a wizard page view.

        :param parent:
            is the optional parent view.
        :return:
            the view which will be an instance of
            :class:`~PyQt4.QtGui.QWizardPage`.
        """

        return QWizardPage(as_QWidget_parent(parent))
