# Copyright (c) 2011 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Instance, List

from .action import Action
from .i_controller import IController
from .dock import Dock
from .i_dock import IDock
from .i_main_window import IMainWindow
from .menu_bar import MenuBar
from .single_subview_container_factory import SingleSubviewContainerFactory
from .i_view_binding import IViewBinding


class MainWindow(SingleSubviewContainerFactory):
    """ The MainWindow class implements a main window factory, i.e. it creates
    views that implement the :class:`~dip.ui.IMainWindow` interface.
    """

    # The list of dock factories.
    docks = List(Dock)

    # The identifier of each widget created by the factory.
    id = 'dip.main_window'

    # The factory for the menu bar.
    menu_bar = Instance(MenuBar)

    # The specific interface implemented by the view.
    view_interface = IMainWindow

    def create_subviews(self, model, controller, view):
        """ Create the sub-views for a containing view.

        :param model:
            is the model.
        :param controller:
            is the controller.
        :param view:
            is the containing view.
        :return:
            the sub-views.
        """

        imainwindow = IMainWindow(view)
        icontroller = IController(controller)

        # Create any docks before any actions they define are referenced.
        for dock_factory in self.docks:
            dock = dock_factory.create(model, controller, view)

            # Configure any action.
            action = IDock(dock).action
            if action is not None:
                icontroller.add_action(action)

            imainwindow.docks.append(dock)

        # Create any menu bar.
        if self.menu_bar is not None:
            imainwindow.menu_bar = self.menu_bar.create(model, controller,
                    view)

        return super().create_subviews(model, controller, view)

    def finalize(self, view):
        """ Reimplemented to finalize any menu bar and docks.

        :param view:
            is the view created by this factory.
        """

        imainwindow = IMainWindow(view)

        # Finalize any menu bar.
        menu_bar = imainwindow.menu_bar
        if menu_bar is not None:
            IViewBinding(menu_bar).factory.finalize(menu_bar)

        # Finalize any docks.
        for dock in imainwindow.docks:
            IViewBinding(dock).factory.finalize(dock)

        super().finalize(view)
