# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Bool, Callable, List, MappingProxy, Model, Str, Subclass
from ..toolkits import Toolkit

from .i_controller import IController
from .object_factory import ObjectFactory
from .i_view import IView
from .i_view_binding import IViewBinding


class ViewFactory(ObjectFactory):
    """ The ViewFactory class is the base class for all factories that
    create views, sub-views and editors.
    """

    # The controller factory.  This must return an object that implements or
    # can be adapted to the :class:`~dip.ui.IController` interface.
    controller_factory = Callable()

    # This is set if the widget is enabled.
    enabled = Bool(None, allow_none=True)

    # The status tip.
    status_tip = Str(None, allow_none=True)

    # The tool tip.
    tool_tip = Str(None, allow_none=True)

    # The specific interface implemented by the view.
    view_interface = Subclass(IView, default=IView)

    # This is set if the widget is visible.
    visible = Bool(None, allow_none=True)

    # The "What's This?" help.
    whats_this = Str(None, allow_none=True)

    # The window title.
    window_title = Str(None, allow_none=True)

    def __call__(self, model=None, controller=None, parent=None, top_level=True, **properties):
        """ Create a view and bind it to a model and optional controller.

        :param model:
            is the optional model which is either a :class:`~dip.model.Model`
            instance or a Python mapping object.
        :param controller:
            is the optional controller.  If this is ``None`` then
            :attr:`~dip.ui.ViewFactory.controller_factory` will be called to
            create one.
        :param parent:
            is the optional parent view.
        :param top_level:
            is ``True`` if the view is to be displayed as a top-level widget.
        :param \*\*properties:
            are any remaining keyword arguments and are interpreted as
            attribute names and values to be set for the view.
        :return:
            the view.
        """

        # Make sure we have a Model for the model.
        if model is None:
            model = Model()
        elif not isinstance(model, Model):
            model = MappingProxy(model)

        # Make sure we have a controller.
        if controller is None:
            # Pass the model rather than set it explicitly to allow the
            # controller to use observe() as a decorator.
            controller = self.controller_factory(model=model)
            icontroller = IController(controller)
        else:
            icontroller = IController(controller)
            icontroller.model = model

        # Create the view itself.
        view = self.create(model, controller, parent)

        icontroller.view = view

        for name, value in properties.items():
            setattr(self.view_interface(view), name, value)

        # Allow further configuration now that everything is created.
        self.finalize(view)

        # Update the view according to the current state of the model.
        icontroller.validate()

        if top_level:
            # Make sure we can display the view.
            view = Toolkit.displayable(view)

        return view

    def create(self, model, controller, parent):
        """ Create the view instance.

        :param model:
            is the model.
        :param controller:
            is the controller.
        :param parent:
            is the optional parent view.
        :return:
            the view.
        """

        view = self.object_for_interface(self.view_interface, parent)
        self.bind_view(view, controller)

        return view

    def bind_view(self, view, controller):
        """ Bind a view created by this factory.

        :param view:
            is the view.
        :param controller:
            is the controller.
        """

        iviewbinding = IViewBinding(view)
        iviewbinding.controller = controller
        iviewbinding.factory = self

        self.configure(view)

    def configure(self, view):
        """ Reimplemented to configure the IView specific attributes.

        :param view:
            is the view to configure.
        """

        iview = IView(view)

        if self.enabled is not None:
            iview.enabled = self.enabled

        if self.status_tip is not None:
            iview.status_tip = self.status_tip

        if self.tool_tip is not None:
            iview.tool_tip = self.tool_tip

        if self.visible is not None:
            iview.visible = self.visible

        if self.whats_this is not None:
            iview.whats_this = self.whats_this

        if self.window_title is not None:
            iview.window_title = self.window_title

        super().configure(view)

    def finalize(self, view):
        """ This is called when the creation fo a view has been completed.
        This default implementation does nothing.

        :param view:
            is the view created by this factory.
        """

    @controller_factory.default
    def controller_factory(self):
        """ Invoked to return the default controller factory. """

        from .controller import Controller

        return Controller
