import sys

from PyQt4.QtCore import QRegExp
from PyQt4.QtGui import QApplication, QRegExpValidator

from dip.model import Enum, Int, Model, Str
from dip.ui import (Dialog, DialogController, IEditor, GroupBox, MessageArea,
        SpinBox)


class Person(Model):
    """ The Person class encapsulates the data held about an individual. """

    # The person's name.
    name = Str(
            validator=lambda ed: QRegExpValidator(QRegExp(r'\w[\w\s]*\w'), ed),
            tool_tip="The person's name",
            whats_this="The name of the person. It must be at least two "
                    "characters and can include spaces.")

    # The person's gender.
    gender = Enum('female', 'male')

    # The person's age.
    age = Int(1, minimum=1)

    # The number of the children the person has had.  It is only applicable if
    # the person is female.
    children = Int()

    # The person's driving license number.  A person must be 17 or older to
    # hold a driving license.
    driving_license = Str(tool_tip="The person's driving license number")


class PersonController(DialogController):
    """ The PersonController class implements a controller for a view that
    allows the update of a Person instance.
    """

    def validate_view(self):
        """ Validate the data in the view.

        :return:
            a string which will be empty if the view is valid, otherwise it
            will explain why the view is invalid.
        """

        # The 'children' editor is only enabled if the gender is female.
        IEditor(self.children_editor).enabled = (IEditor(self.gender_editor).value == 'female')

        # The super-class will handle the basic editor validation.
        invalid_reason = super(PersonController, self).validate_view()

        if invalid_reason == '':
            if IEditor(self.driving_license_editor).value != '' and \
                    IEditor(self.age_editor).value < 17:
                invalid_reason = "A person must be at least 17 years old " \
                        "to hold a driving license."

        return invalid_reason


# Every PyQt GUI application needs a QApplication.
app = QApplication(sys.argv)

# Create the model.
model = Person()

# Define the view.
view = Dialog(
        GroupBox(
            'name',
            SpinBox('age', suffix=" years"),
            'gender',
            'children',
            'driving_license',
            title="Person Record"
        ),
        MessageArea(),
        window_title="Validate"
    )

# Create the controller.
controller = PersonController()

# Create an instance of the view bound to the model.
ui = view(model, controller=controller)

# Enter the dialog's modal event loop.
ui.exec_()

# Show the attributes of the model.
print("Name:", model.name)
print("Gender:", model.gender)
print("Age:", model.age)
print("Children:", model.children)
print("Driving license:", model.driving_license)
